.memtemplate {
    display: none;
}

.memTemplParams {
    display: none;
}

.navtab {
    display: none;
}

#top, .footer {
    display: none;
}

td.paramname {
    vertical-align: top;
}

.ok_green {
    background-color: #89C35C;
}

.nok_throws {
    background-color: #ffa500;
}
