/* hide lengthy template information */
.memtemplate, .memTemplParams {
    display: none;
}

/* allow compiler information to wrap */
/* https://css-tricks.com/snippets/css/make-pre-text-wrap/ */
pre.fragment {
    white-space: pre-wrap;       /* css-3 */
    white-space: -moz-pre-wrap;  /* Mozilla, since 1999 */
    white-space: -pre-wrap;      /* Opera 4-6 */
    white-space: -o-pre-wrap;    /* Opera 7 */
    word-wrap: break-word;       /* Internet Explorer 5.5+ */
}

td.paramname {
    vertical-align: top;
}

.ok_green {
    background-color: #89C35C;
}

.nok_throws {
    background-color: #ffa500;
}
